//
// Parallax occlusion mapping with soft shadows in OpenGL via GLSL
//
// Author: Alex V. Boreskoff <steps3d@gmail.com>, <steps3d@narod.ru>
//

#include    "libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "Vector4D.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include    "GlslProgram.h"

#define CAPTION     "OpenGL parallax occlusion mapping with soft shadows"

Vector3D    eye   ( 7, 5, 7 );          // camera position
Vector3D    light ( 4, 0, 3 );          // light position
unsigned	decalMap;                   // decal map
unsigned	heightMap;                  // height texture
unsigned    bumpMap;
float      	angle = 0;
Vector3D    rot ( 0, 0, 0 );
int         mouseOldX = 0;
int         mouseOldY = 0;
bool        parallax = true;

GlslProgram     parallaxProgram;
GlslProgram		specularProgram;

void	drawFace ()
{
	float size = 5;

	glBegin     ( GL_QUADS );
	glNormal3fv ( Vector3D ( 0, 0, 1 ) );
	glMultiTexCoord3fv ( GL_TEXTURE1_ARB, Vector3D ( 1, 0, 0 ) );
	glMultiTexCoord3fv ( GL_TEXTURE2_ARB, Vector3D ( 0, 1, 0 ) );

	glMultiTexCoord2fv ( GL_TEXTURE0_ARB, Vector2D ( 0, 0 ) );
	glVertex3fv         ( Vector3D ( -size, -size, 0 ) );

	glMultiTexCoord2fv ( GL_TEXTURE0_ARB, Vector2D ( 1, 0 ) );
	glVertex3fv         ( Vector3D ( size, -size, 0 ) );

	glMultiTexCoord2fv ( GL_TEXTURE0_ARB, Vector2D ( 1, 1 ) );
	glVertex3fv         ( Vector3D ( size, size, 0 ) );

	glMultiTexCoord2fv ( GL_TEXTURE0_ARB, Vector2D ( 0, 1 ) );
	glVertex3fv         ( Vector3D ( -size, size, 0 ) );

	glEnd ();
}

void    setParallax ( bool value )
{
    char    buf [256];

    parallax = value;

    sprintf ( buf, "%s - parallax is %s", CAPTION, parallax ? "on" : "off" );

    glutSetWindowTitle ( buf );
}

void init ()
{
    glClearColor ( 0.0, 0.0, 0.0, 1.0 );
    glEnable     ( GL_DEPTH_TEST );
    glDepthFunc  ( GL_LEQUAL );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void display ()
{
    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

                                                // setup texture units
                                                // bind bump (normal) map to texture unit 0
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, decalMap );

    glActiveTextureARB ( GL_TEXTURE1_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, heightMap );

    glActiveTextureARB ( GL_TEXTURE2_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, bumpMap );

    glMatrixMode   ( GL_MATRIX0_ARB );
    glLoadIdentity ();
    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );

    glMatrixMode   ( GL_MODELVIEW );
    glPushMatrix   ();

    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );


    if ( parallax )
        parallaxProgram.bind ();
    else
        specularProgram.bind  ();

	drawFace ();

	if ( parallax )
	    parallaxProgram.unbind ();
	else
		specularProgram.unbind ();

    glPopMatrix     ();
    glutSwapBuffers ();
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( eye.x, eye.y, eye.z,    // eye
                    0, 0, 0,                // center
                    0, 0, 1 );              // up
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
        rot.z -= 360;

    if ( rot.z < -360 )
        rot.z += 360;

    if ( rot.y > 360 )
        rot.y -= 360;

    if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

    glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
    }
}

void key ( unsigned char key, int x, int y )
{
                                // quit requested
    if ( key == 27 || key == 'q' || key == 'Q' )
        exit ( 0 );

    if ( key == 'p' || key == 'P' )
        setParallax ( !parallax );
}

void    animate ()
{
    angle  = 3*0.001f * glutGet ( GLUT_ELAPSED_TIME );

    light.x = 4+7*cos ( angle );
    light.y = 4+6*sin ( angle );
    light.z = 3;// + sin ( angle / 3 );

    parallaxProgram.bind ();
    parallaxProgram.setUniformVector ( "eye",   Vector4D ( eye,   1 )  );
    parallaxProgram.setUniformVector ( "light", Vector4D ( light, 1 ) );
    parallaxProgram.unbind ();

    specularProgram.bind ();
    specularProgram.setUniformVector ( "eye",   Vector4D ( eye,   1 ) );
    specularProgram.setUniformVector ( "light", Vector4D ( light, 1 ) );
    specularProgram.unbind ();

    glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    glutInitWindowSize  ( 500, 500 );


                                // create window
    glutCreateWindow ( CAPTION );

                                // register handlers
    glutDisplayFunc  ( display );
    glutReshapeFunc  ( reshape );
    glutKeyboardFunc ( key     );
    glutMouseFunc    ( mouse   );
    glutMotionFunc   ( motion  );
    glutIdleFunc     ( animate );

    init           ();
    initExtensions ();

    decalMap  = createTexture2D ( true,  "Maps/relief_wood.jpg" );
    heightMap = createTexture2D ( true,  "Maps/relief_tile1.tga" );

	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported.\n" );

		return 1;
	}

    if ( !specularProgram.loadShaders ( "specular-5.vsh", "specular-5.fsh" ) )
    {
        printf ( "Error loading specular shaders:\n%s\n", specularProgram.getLog ().c_str () );

        return 3;
    }

    if ( !parallaxProgram.loadShaders ( "parallax-occlusion-3.vsh", "parallax-occlusion-3.fsh" ) )
    {
        printf ( "Error loading parallax shaders:\n%s\n", parallaxProgram.getLog ().c_str () );

        return 2;
    }

	printf ( "Press P to toggle effect\n" );
	
    parallaxProgram.bind ();
    parallaxProgram.setTexture ( "decalMap",  0 );
    parallaxProgram.setTexture ( "heightMap", 1 );
    parallaxProgram.unbind ();

    specularProgram.bind ();
    specularProgram.setTexture ( "decalMap",  0 );
    specularProgram.setTexture ( "heightMap", 1 );
    specularProgram.unbind ();

    setParallax  ( true );
    glutMainLoop ();

    return 0;
}
