//
// Sample to parallax mapping in OpenGL
//
// Author: Alex V. Boreskoff <steps3d@gmail.com>, <steps3d@narod.ru>
//

#include    "libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include    "GlslProgram.h"

#define CAPTION     "OpenGL parallax mapping"

Vector3D    eye   ( 7, 5, 7 );          // camera position
Vector3D    light ( 5, 0, 4 );          // light position
unsigned    decalMap;                   // decal map
unsigned    heightMap;                  // height texture
unsigned    bumpMap;
Vector3D    rot ( 0, 0, 0 );
float       angle     = 0;
int         mouseOldX = 0;
int         mouseOldY = 0;
bool        parallax  = true;
float	  	scale     = 0.03;
float	    bias      = 0.015;

GlslProgram     program;

void	drawFace ()
{
	float size = 5;

	glBegin     ( GL_QUADS );
	glNormal3fv ( Vector3D ( 0, 0, 1 ) );
	glMultiTexCoord3fv ( GL_TEXTURE1_ARB, Vector3D ( 1, 0, 0 ) );
	glMultiTexCoord3fv ( GL_TEXTURE2_ARB, Vector3D ( 0, 1, 0 ) );

	glMultiTexCoord2fv ( GL_TEXTURE0_ARB, Vector2D ( 0, 0 ) );
	glVertex3fv         ( Vector3D ( -size, -size, 0 ) );

	glMultiTexCoord2fv ( GL_TEXTURE0_ARB, Vector2D ( 1, 0 ) );
	glVertex3fv         ( Vector3D ( size, -size, 0 ) );

	glMultiTexCoord2fv ( GL_TEXTURE0_ARB, Vector2D ( 1, 1 ) );
	glVertex3fv         ( Vector3D ( size, size, 0 ) );

	glMultiTexCoord2fv ( GL_TEXTURE0_ARB, Vector2D ( 0, 1 ) );
	glVertex3fv         ( Vector3D ( -size, size, 0 ) );

	glEnd ();
}

void    setParallax ( bool value )
{
    char    buf [256];

    parallax = value;

    sprintf ( buf, "%s - parallax is %s", CAPTION, parallax ? "on" : "off" );

    glutSetWindowTitle ( buf );
}

void init ()
{
    glClearColor ( 0.0, 0.0, 0.0, 1.0 );
    glEnable     ( GL_DEPTH_TEST );
    glDepthFunc  ( GL_LEQUAL );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void display ()
{
    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

                                                // setup texture units
                                                // bind bump (normal) map to texture unit 0
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, decalMap );

    glActiveTextureARB ( GL_TEXTURE1_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, heightMap );

    glActiveTextureARB ( GL_TEXTURE2_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, bumpMap );

    glMatrixMode   ( GL_MATRIX0_ARB );
    glLoadIdentity ();
    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );

    glMatrixMode   ( GL_MODELVIEW );
    glPushMatrix   ();

    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );


    if ( parallax )
    {
        program.bind ();
    }
    else
    {
        program.unbind  ();

        glActiveTextureARB ( GL_TEXTURE0_ARB );
        glEnable           ( GL_TEXTURE_2D );
        glTexEnvi          ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
        glActiveTextureARB ( GL_TEXTURE1_ARB );
        glDisable          ( GL_TEXTURE_2D );
        glActiveTextureARB ( GL_TEXTURE2_ARB );
        glDisable          ( GL_TEXTURE_2D );
    }

	drawFace ();

    program.unbind ();

    glPopMatrix     ();
    glutSwapBuffers ();
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( eye.x, eye.y, eye.z,    // eye
                    0, 0, 0,                // center
                    0, 0, 1 );              // up
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
        rot.z -= 360;

    if ( rot.z < -360 )
        rot.z += 360;

    if ( rot.y > 360 )
        rot.y -= 360;

    if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

    glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
    }
}

void key ( unsigned char key, int x, int y )
{
                                // quit requested
    if ( key == 27 || key == 'q' || key == 'Q' )
        exit ( 0 );

    if ( key == 'p' || key == 'P' )
        setParallax ( !parallax );
	else
	if ( key == '+' )
		scale += 0.0005;
	else
	if ( key == '-' )
		scale -= 0.0005;
}

void    animate ()
{
    angle  = 0.001f * glutGet ( GLUT_ELAPSED_TIME );

    light.x = 3.5*cos ( angle );
    light.y = sin ( angle );
    light.z = 4 + sin ( angle / 3 );

    if ( parallax )
    {
        program.bind ();
        program.setUniformVector ( "eye",   eye   );
        program.setUniformVector ( "light", light );
		program.setUniformFloat  ( "scale", scale );
		program.setUniformFloat  ( "bias",  bias  );
        program.unbind ();
    }

    glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    glutInitWindowSize  ( 500, 500 );


                                // create window
    glutCreateWindow ( CAPTION );

                                // register handlers
    glutDisplayFunc  ( display );
    glutReshapeFunc  ( reshape );
    glutKeyboardFunc ( key     );
    glutMouseFunc    ( mouse   );
    glutMotionFunc   ( motion  );
    glutIdleFunc     ( animate );

    init           ();
    initExtensions ();

	if ( argc > 1 && !strcmp ( argv [1], "wood" ) )
	{
	    decalMap  = createTexture2D ( true,  "Maps/relief_wood.jpg" );
	    heightMap = createTexture2D ( true,  "Maps/relief_height.png" );
	    bumpMap   = createTexture2D ( false, "Maps/relief_bump.png" );
	}
	else
	{
	    decalMap  = createTexture2D ( true,  "Maps/rockwall.tga" );
	    heightMap = createTexture2D ( true,  "Maps/rockwall_height.tga" );
	    bumpMap   = createTexture2D ( false, "Maps/rockwall_normal.tga" );
	}

	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported.\n" );

		return 1;
	}

    if ( !program.loadShaders ( "parallax.vsh", "parallax.fsh" ) )
    {
        printf ( "Error loading shaders:\n%s\n", program.getLog ().c_str () );

        return 3;
    }

	printf ( "Press P to toggle effect\nUse + and - keys to change parallax scale\n" );
	
    program.bind ();
    program.setTexture ( "decalMap",  0 );
    program.setTexture ( "heightMap", 1 );
    program.unbind ();

    setParallax  ( true );
    glutMainLoop ();

    return 0;
}
